import filecmp
import glob
import json
import os
import re
import shutil
import subprocess
import sys
import tempfile
import time

# Various bool functions to help identify the host environment
def is_redhat_based() -> bool:
    return os.path.exists('/etc/redhat-release')

def is_debian_based() -> bool:
    return os.path.exists('/etc/debian_version')

def is_arch_linux() -> bool:
    return os.path.exists('/etc/arch-release')

def is_ubuntu() -> bool:
    if not is_debian_based():
        return False
    try:
        with open('/etc/os-release') as f:
            for line in f:
                if line.startswith('ID=ubuntu'):
                    return True
    except Exception:
        pass
    return False

def is_linux() -> bool:
    return is_debian_based() or is_redhat_based() or is_arch_linux()

def is_macos() -> bool:
    return sys.platform == 'darwin'

def is_windows() -> bool:
    return sys.platform == 'win32'

def is_cmake_installed() -> bool:
    try:
        subprocess.run(['cmake', '--version'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except subprocess.CalledProcessError:
        return False
    return True

def is_rpmbuild_installed() -> bool:
    if not is_redhat_based():
        return False
    try:
        subprocess.run(['rpmbuild', '--version'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except subprocess.CalledProcessError:
        return False
    return True

def is_dpkg_installed() -> bool:
    if not is_debian_based():
        return False
    try:
        subprocess.run(['dpkg', '--version'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except subprocess.CalledProcessError:
        return False
    return True

def is_dotnet_installed() -> bool:
    try:
        subprocess.run(['dotnet', '--version'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return True
    except (subprocess.CalledProcessError, FileNotFoundError):
        return False

def is_wix_installed() -> bool:
    # For installation, see https://cmake.org/cmake/help/latest/cpack_gen/wix.html#wix-net-tools
    try:
        subprocess.run(['wix', '--version'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return True
    except (subprocess.CalledProcessError, FileNotFoundError):
        return False

def is_msbuild_installed() -> bool:
    if sys.platform == "win32":
        try:
            vswhere_path = r"C:\Program Files (x86)\Microsoft Visual Studio\Installer\vswhere.exe"
            if not os.path.exists(vswhere_path):
                return False                

            result = subprocess.run([vswhere_path, '-latest', '-products', '*', '-requires', 'Microsoft.Component.MSBuild', '-find', 'MSBuild\\**\\Bin\\MSBuild.exe'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            msbuild_path = result.stdout.decode().strip()
            if msbuild_path:
                subprocess.run([msbuild_path, '-version'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                return True
        except (subprocess.CalledProcessError, FileNotFoundError):
            return False
    return False

def is_brew_installed() -> bool:
    try:
        subprocess.run(['brew', '--version'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return True
    except (subprocess.CalledProcessError, FileNotFoundError):
        return False
    return False

def is_arm64_toolchain_installed() -> bool:
    if is_linux():
        try:
            subprocess.run(['aarch64-linux-gnu-gcc', '--version'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            return True
        except (subprocess.CalledProcessError, FileNotFoundError):
            return False
    elif is_windows:
        if not is_msbuild_installed():
            return False

    # TODO - Add tool specific detection for Windows/MacOS

    return sys.platform.machine().lower() in ['aarch64', 'arm64']

def is_x86_64_toolchain_installed() -> bool:
    if is_linux():
        try:
            subprocess.run(['x86_64-linux-gnu-gcc', '--version'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            return True
        except (subprocess.CalledProcessError, FileNotFoundError):
            return False
    elif is_windows:
        if not is_msbuild_installed():
            return False

    # TODO - Add more tool specific detection for Windows/MacOS

    return sys.platform.machine().lower() in ['amd64', 'x86_64']

def is_i386_toolchain_installed() -> bool:
    if is_linux():
        try:
            subprocess.run(['i686-linux-gnu-gcc', '--version'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            return True
        except (subprocess.CalledProcessError, FileNotFoundError):
            return False

    # TODO - Add tool specific detection for Windows/MacOS
    return sys.platform.machine().lower() in ['i386', 'i686']

# Utility function to list all assets expected for a release.
def get_release_assets(version:str) -> list:
    """
    Return the list of assets expected for a release.

    This is used for CI.

    TA-Lib maintainers should modify this list everytime an assets is added/removed
    """
    return [
        f'ta-lib-{version}-src.tar.gz',
        f'ta-lib-{version}-windows-x86_64.msi',
        f'ta-lib-{version}-windows-x86_64.zip',
        f'ta-lib-{version}-windows-x86_32.msi',
        f'ta-lib-{version}-windows-x86_32.zip',
        f'ta-lib_{version}_amd64.deb',
        f'ta-lib_{version}_arm64.deb',
        f'ta-lib_{version}_i386.deb',
    ]

# Utility functions to identify the gen_code generated files.
def get_src_generated_files() -> list:
    """
    Return the list of generated files and directories.

    This is only for what is expected in the src.tar.gz package.

    Everything under a directory ('**') and file glob allowed ('*')

    See get_all_generated_files() for more...

    This is used for CI.

    TA-Lib maintainers should update this list everytime a new file is generated (or not).
    """
    return [
        'include/ta_func.h',
        'include/ta_defs.h',
        'src/ta_func/*.c',
        'src/ta_abstract/*.c',
        'src/ta_abstract/frames/*.c',
        'src/ta_abstract/frames/*.h',
        'src/ta_common/ta_retcode.c',
        'src/ta_abstract/ta_java_defs.h',
    ]

def get_all_generated_files() -> list:
    """
    Returns list of all generated files and directories.
    Everything under a directory ('**') and file glob allowed ('*')

    This is used for CI.

    TA-Lib maintainers should update this list everytime a new file is generated (or not).
    """
    return [
        'swig/src/interface/ta_func.swg',
        'dotnet/src/Core/TA-Lib-Core.vcproj',
        'dotnet/src/Core/TA-Lib-Core.h',
        'ide/msvc/lib_proj/ta_func/ta_func.dsp',
        'java/src/**',
    ]  + get_src_generated_files()

def expand_globs(root_dir: str, file_list: list) -> list:
    """
    Expand glob patterns in the file list to actual file paths.
    """
    expanded_files = []
    for file in file_list:
        # Use recursive globbing if '**' is in the pattern
        if '**' in file:
            expanded_files.extend(glob.glob(os.path.join(root_dir, file), recursive=True))
        else:
            expanded_files.extend(glob.glob(os.path.join(root_dir, file)))
    return expanded_files


def run_command(command: list) -> str:
    """Run a shell command and return the output."""
    try:
        result = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        if result.returncode != 0:
            print(f"stdout for '{' '.join(command)}': {result.stdout}")
            print(f"stderr for '{' '.join(command)}': {result.stderr}")
            # If result.stderr contains the string "CPack Error: Problem running WiX.", then
            # print the content of the log file at the specified path.
            if "CPack Error: Problem running WiX." in result.stderr:
                log_file_match = re.search(r"'([^']+)'", result.stderr)
                if log_file_match:
                    log_file = log_file_match.group(1)
                    if os.path.exists(log_file):
                        print(f"Contents of {log_file}:")
                        with open(log_file, 'r') as f:
                            print(f.read())
            sys.exit(1)
    except subprocess.CalledProcessError as e:
        print(f"Error during '{' '.join(command)}': {e}")
        sys.exit(1)

    return result.stdout.strip()

def run_command_term(command: list):
    """
    Similar to run_command, but let it print its output to the terminal instead of capturing.
    Also, exit if the command fails.
    """
    try:
        subprocess.run(command, check=True)
    except subprocess.CalledProcessError as e:
        print(f"Error during '{' '.join(command)}': {e}")
        sys.exit(1)

def run_command_sudo(command: list, sudo_pwd: str=''):
    """
    Run a command with sudo, optionally using a password if provided.
    Will exit the script if calling the command fails or exit code != 0.
    """
    try:
        if sudo_pwd:
            # Pipeline the password to sudo
            process = subprocess.Popen(['sudo', '-S'] + command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, stderr = process.communicate(input=f'{sudo_pwd}\n'.encode())
            if process.returncode != 0:
                print(f"Error during 'sudo {' '.join(command)}': {stderr.decode()}")
                sys.exit(1)
            return stdout.decode().strip()
        else:
            return run_command(['sudo'] + command)
    except subprocess.CalledProcessError as e:
        print(f"Error during 'sudo {' '.join(command)}': {e}")
        sys.exit(1)
    except Exception as e:
        print(f"Unexpected error during 'sudo {' '.join(command)}': {e}")
        sys.exit(1)

    return f"Unexpected error: {' '.join(command)}"


def create_temp_dir(root_dir) -> str:
    # Create a temporary directory under root_dir/temp, also purge older ones.
    #
    # Return the path of the newly created directory.

    # Delete oldest directories if more than 10 exists and it is more
    # than 1 hour old.
    temp_root_dir = os.path.join(root_dir, "temp")
    os.makedirs(temp_root_dir, exist_ok=True)
    temp_dirs = sorted(os.listdir(temp_root_dir), key=lambda x: os.path.getctime(os.path.join(temp_root_dir, x)))
    if len(temp_dirs) > 10:
        for i in range(len(temp_dirs) - 10):
            temp_dir_path = os.path.join(temp_root_dir, temp_dirs[i])
            if os.path.isdir(temp_dir_path) and (time.time() - os.path.getctime(temp_dir_path)) > 3600:
                shutil.rmtree(temp_dir_path)

    # Create the new temp directory
    return tempfile.mkdtemp(dir=temp_root_dir)

def verify_git_repo() -> str:
    # Verify that the script is called from within a ta-lib git repos, and if yes
    # change the working directory to the root of it.
    #
    # That root path is returned.
    try:
        subprocess.run(['git', '--version'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except subprocess.CalledProcessError:
        print("Git is not installed. Please install Git and try again.")
        sys.exit(1)

    error = False
    try:
        result = subprocess.run(['git', 'rev-parse', '--is-inside-work-tree'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if result.stdout.strip() == b'true':
            # Change to the root directory of the Git repository
            root_dir = subprocess.run(['git', 'rev-parse', '--show-toplevel'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE).stdout.strip().decode('utf-8')
            os.chdir(root_dir)
            return root_dir
        else:
            error = True
    except subprocess.CalledProcessError:
        error = True

    if error:
        print("Must run this script while the current directory is in a TA-Lib Git repository.")
        sys.exit(1)

    # Sanity check that src/ta_func exists.
    if not os.path.isdir('src/ta_func'):
        print("Current directory is not a complete TA-Lib Git repository (src/ta_func missing)")
        sys.exit(1)

def verify_git_repo_original() -> str:
    # Similar to verify_git_repo, but checks additionally if running from
    # the original repos (will return false if a fork).
    #
    # Some operations (e.g. publishing the official TA-Lib on brew) should
    # be attempted to be done exclusively with the original repos.
    #
    root_dir = verify_git_repo()

    try:
        # Get the remote URL of the repository
        result = subprocess.run(
            ["git", "config", "--get", "remote.origin.url"],
            capture_output=True,
            text=True,
            check=True
        )
        remote_url = result.stdout.strip()

        # Normalize the URL to a common format (account for ssh and https differences).
        normalized_url = re.sub(r'^git@github\.com:', 'https://github.com/', remote_url)
        normalized_url = re.sub(r'\.git$', '', normalized_url).lower()

        # print(f"Remote URL: {normalized_url}")
        # Check if the normalized URL matches the original ta-lib repository URL
        if normalized_url.endswith("ta-lib/ta-lib"):
            return root_dir

    except subprocess.CalledProcessError:
        print("Error: Could not determine the remote URL of the repository.")
        sys.exit(1)

    print("Warning: some processing were skipped because running from a fork")
    sys.exit(0)

def get_git_bot_user_name() -> str:
    return "github-actions[bot]"

def get_git_user_name() -> str:
    if os.getenv('GITHUB_ACTIONS') == 'true':
        return get_git_bot_user_name()

    try:
        result = subprocess.run(['git', 'config', 'user.name'], capture_output=True, text=True, check=True)
        user_name = result.stdout.strip()
        if not user_name:
            user_name = "local"
        return user_name
    except subprocess.CalledProcessError as e:
        return "local"

def is_nightly_github_action() -> bool:
    if os.getenv('GITHUB_ACTIONS') != 'true':
        return False

    event_name = os.getenv('GITHUB_EVENT_NAME')
    if event_name != 'schedule':
        return False

    event_path = os.getenv('GITHUB_EVENT_PATH')
    if not event_path or not os.path.exists(event_path):
        return False

    with open(event_path, 'r') as f:
        event_data = json.load(f)

    # Check if the event was manually dispatched
    if event_data.get('action') == 'workflow_dispatch':
        return False

    return True

def are_generated_files_git_changed(root_dir: str) -> bool:
    # Using git, verify if any of the generated files have changed.
    #
    # root_dir must be the root of the TA-Lib Git repository.
    original_dir = os.getcwd()
    os.chdir(root_dir)

    try:
        result = subprocess.run(['git', 'diff', '--exit-code', 'HEAD', '--'] + get_all_generated_files(), check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return result.returncode != 0

    except subprocess.CalledProcessError as e:
        print(f"Error: {e}")
        return False

    finally:
        os.chdir(original_dir)
    return False

def copy_file_list(src_dir: str, dest_dir: str, file_list: list):
    # Copy the files and directory to dest_dir.
    #
    # The file list can include whole directories ('**') and file glob ('*').
    #
    # 'dest_dir' can then be used later to detect changes
    # with compare_dir.

    # Delete all contents in dest_dir, but not dest_dir itself.
    if os.path.exists(dest_dir):
        for root, dirs, files in os.walk(dest_dir):
            for file in files:
                os.remove(os.path.join(root, file))
            for dir in dirs:
                shutil.rmtree(os.path.join(root, dir))

    os.makedirs(dest_dir, exist_ok=True)

    expanded_files = expand_globs(src_dir, file_list)
    for src_file in expanded_files:
        dest_file = os.path.join(dest_dir, os.path.relpath(src_file, src_dir))
        os.makedirs(os.path.dirname(dest_file), exist_ok=True)
        if os.path.isdir(src_file):
            shutil.copytree(src_file, dest_file, dirs_exist_ok=True)
        else:
            shutil.copy(src_file, dest_file)

def compare_dir(dir1: str, dir2: str) -> bool:
    # Detect any difference in files or directory.
    # For files, also does a binary comparison.
    # Recursively check subdirectories.
    dircmp = filecmp.dircmp(dir1, dir2)

    differences_found = False

    if dircmp.left_only:
        print(f"Files only in {dir1}: {dircmp.left_only}")
        differences_found = True

    if dircmp.right_only:
        print(f"Files only in {dir2}: {dircmp.right_only}")
        differences_found = True

    if dircmp.diff_files:
        print(f"Files that differ: {dircmp.diff_files}")
        differences_found = True

    if dircmp.funny_files:
        print(f"Files that could not be compared: {dircmp.funny_files}")
        differences_found = True

    for subdir in dircmp.common_dirs:
        if not compare_dir(os.path.join(dir1, subdir), os.path.join(dir2, subdir)):
            differences_found = True

    return not differences_found




import hashlib
import re
from typing import Union

import glob
import os
import subprocess
import sys
import zipfile
import tempfile
import filecmp
import zlib
import tarfile

from .common import is_linux, is_windows, run_command, run_command_sudo

# Use path_join to create a new path with proper seperators for the host system.
# (this is to solve a problem on windows with '\' versus '/')
def path_join(*args: Union[str, os.PathLike]) -> str:
    return os.path.normpath(os.path.join(*args))

def compare_dir_recursive(dir1, dir2):
    # Recursively compare subdirectories and files content
    comparison = filecmp.dircmp(dir1, dir2)
    if comparison.diff_files:
        print(f"Differing files: {comparison.diff_files}")
    if comparison.left_only:
        print(f"Files only in {dir1}: {comparison.left_only}")
    if comparison.right_only:
        print(f"Files only in {dir2}: {comparison.right_only}")
    result = not comparison.diff_files and not comparison.left_only and not comparison.right_only
    if not result:
        return False

    for subdir in comparison.common_dirs:
        if not compare_dir_recursive(path_join(dir1, subdir), path_join(dir2, subdir)):
            return False

    return True # Identical

def remove_lib_files_recursive(target_path: str):
    # Comparing libraries is often not deterministic even if the source code was not changed
    # (e.g. because of embedded __DATE__).
    #
    # This function remove all library files at the target_path, leaving
    # presumably only the source files.
    #
    # Note: To further help detect differences in any source files, all
    # packages should includes ta_common.h which provides a source digest.
    #
    # That source digest is necessary because some package includes only the
    # headers... not all the files that were used to build the package (but
    # all are reflected in the digest)
    #
    # Always remove the ta_config.h... it is a build artifact.
    if is_windows():
        for root, dirs, files in os.walk(target_path):
            for file in files:
                if "ta_config.h" in file or file.endswith('.msi') or file.endswith('.lib') or file.endswith('.dll'):
                    os.remove(path_join(root, file))
    elif is_linux():
        pattern = r'\.so\.\d+'
        for root, dirs, files in os.walk(target_path):
            for file in files:
                if "ta_config.h" in file or file.endswith('.a') or file.endswith('.so') or re.search(pattern, file) is not None:
                    os.remove(path_join(root, file))

    



def compare_zip_files(zip_file1, zip_file2):
    # Does a comparison of the contents of the two zip files.
    # Ignores file creation time and binaries.
    with tempfile.TemporaryDirectory() as temp_extract_dir:
        temp_extract_path1 = path_join(temp_extract_dir, 'temp1')
        temp_extract_path2 = path_join(temp_extract_dir, 'temp2')
        os.makedirs(temp_extract_path1, exist_ok=True)
        os.makedirs(temp_extract_path2, exist_ok=True)

        with zipfile.ZipFile(zip_file1, 'r') as zip_ref:
            zip_ref.extractall(temp_extract_path1)

        with zipfile.ZipFile(zip_file2, 'r') as zip_ref:
            zip_ref.extractall(temp_extract_path2)

        # Remove in both temp directory all library files.
        # Only the remaining source files will be compared.
        remove_lib_files_recursive(temp_extract_path1)
        remove_lib_files_recursive(temp_extract_path2)

        return compare_dir_recursive(temp_extract_path1, temp_extract_path2)

def create_zip_file(source_dir, zip_file):
    with zipfile.ZipFile(zip_file, 'w', compression=zipfile.ZIP_DEFLATED, compresslevel=zlib.Z_BEST_COMPRESSION) as zipf:
        for root, dirs, files in os.walk(source_dir):
            for file in files:
                file_path = path_join(root, file)
                arcname = os.path.relpath(file_path, start=source_dir)
                zipf.write(file_path, arcname)

def compare_tar_gz_files(tar_gz_file1, tar_gz_file2) -> bool:
    # Does a binary comparison of the contents of the two tar.gz files.
    # Ignores file creation time.
    with tempfile.TemporaryDirectory() as temp_extract_dir:
        temp_extract_path1 = os.path.join(temp_extract_dir, 'temp1')
        temp_extract_path2 = os.path.join(temp_extract_dir, 'temp2')
        os.makedirs(temp_extract_path1, exist_ok=True)
        os.makedirs(temp_extract_path2, exist_ok=True)

        with tarfile.open(tar_gz_file1, 'r:gz') as tar_ref:
            tar_ref.extractall(temp_extract_path1)

        with tarfile.open(tar_gz_file2, 'r:gz') as tar_ref:
            tar_ref.extractall(temp_extract_path2)

        # Remove in both temp directory all library files.
        # Only the remaining source files will be compared.
        remove_lib_files_recursive(temp_extract_path1)
        remove_lib_files_recursive(temp_extract_path2)

        # Because of subtle difference between autotools versions,
        # ignore also a few of its generated files. They
        # just "do not matter" as much.
        #
        # Source file differences are what matter the most
        # to trig re-packaging.
        for root, dirs, files in os.walk(temp_extract_path1):
            for file in files:
                if file == 'Makefile.in' or file == 'configure' or file == 'ltmain.sh' or file.endswith('.m4'):
                    os.remove(os.path.join(root, file))

        for root, dirs, files in os.walk(temp_extract_path2):
            for file in files:
                if file == 'Makefile.in' or file == 'configure' or file == 'ltmain.sh'  or file.endswith('.m4'):
                    os.remove(os.path.join(root, file))

        return compare_dir_recursive(temp_extract_path1, temp_extract_path2)

def create_tar_gz_file(source_dir, tar_gz_file):
    with tarfile.open(tar_gz_file, 'w:gz') as tarf:
        for root, dirs, files in os.walk(source_dir):
            for file in files:
                file_path = os.path.join(root, file)
                arcname = os.path.relpath(file_path, start=source_dir)
                tarf.add(file_path, arcname)


def compare_deb_files(deb_file1, deb_file2) -> bool:
    # Does a binary comparison of the contents of the two .deb files.
    # Ignores file creation time and binaries.
    with tempfile.TemporaryDirectory() as temp_extract_dir:

        temp_extract_path1 = os.path.join(temp_extract_dir, 'temp1')
        temp_extract_path2 = os.path.join(temp_extract_dir, 'temp2')
        os.makedirs(temp_extract_path1, exist_ok=True)
        os.makedirs(temp_extract_path2, exist_ok=True)

        try:
            subprocess.run(['dpkg-deb', '-x', deb_file1, temp_extract_path1], check=True)
            subprocess.run(['dpkg-deb', '-x', deb_file2, temp_extract_path2], check=True)
        except subprocess.CalledProcessError as e:
            print(f"Error extracting .deb files: {e}")
            return False

        # Remove in both temp directory all library files.
        # Only the remaining source files will be compared.
        remove_lib_files_recursive(temp_extract_path1)
        remove_lib_files_recursive(temp_extract_path2)

        return compare_dir_recursive(temp_extract_path1, temp_extract_path2)

def compare_msi_files(msi_file1, msi_file2) -> bool:
    # Does a binary comparison of the contents of the two .msi files.
    # Ignores file creation time and binaries
    with tempfile.TemporaryDirectory() as temp_extract_dir:
        temp_extract_path1 = path_join(temp_extract_dir, 'temp1')
        temp_extract_path2 = path_join(temp_extract_dir, 'temp2')
        os.makedirs(temp_extract_path1, exist_ok=True)
        os.makedirs(temp_extract_path2, exist_ok=True)
        os.system(f"msiexec /a {msi_file1} /qn TARGETDIR={temp_extract_path1}")
        os.system(f"msiexec /a {msi_file2} /qn TARGETDIR={temp_extract_path2}")

        # Remove in both temp directory all library files.
        # Only the remaining source files will be compared
        remove_lib_files_recursive(temp_extract_path1)
        remove_lib_files_recursive(temp_extract_path2)

        return compare_dir_recursive(temp_extract_path1, temp_extract_path2)

def create_rtf_from_txt(input_file: str, output_file: str):
    """
    Converts a plain text file to an RTF file.

    Args:
        input_file (str): Path to the input plain text file.
        output_file (str): Path to the output RTF file.
    """
    # Create the RTF header
    rtf_header = '{\\rtf1\\ansi\\deff0\n'
    rtf_footer = '}'

    try:
        with open(input_file, 'r') as txt_file, open(output_file, 'w') as rtf_file:
            # Write the RTF header
            rtf_file.write(rtf_header)

            # Add the content of the plain text file, escaping special characters
            for line in txt_file:
                # Escape backslashes, braces, and newlines
                escaped_line = line.replace('\\', '\\\\').replace('{', '\\{').replace('}', '\\}')
                rtf_file.write(escaped_line + '\\par\n')

            # Write the RTF footer
            rtf_file.write(rtf_footer)

        print(f"RTF file created: {output_file}")
    except Exception as e:
        print(f"Error creating RTF file: {e}")

def force_delete(path: str, sudo_pwd: str = ""):
    # Force delete a file or directory.
    #
    # Try first as normal user. On failure, try again as sudo.
    #
    # 'sudo' is sometimes necessary for files that were created as part
    # of the packaging/installation process.
    #
    # Exit on any error.
    try:
        if is_windows():
            if os.path.isdir(path):
                subprocess.run(['cmd', '/c', 'rmdir', '/s', '/q', path], check=True, stderr=subprocess.DEVNULL)
            else:
                subprocess.run(['cmd', '/c', 'del', '/f', '/q', path], check=True, stderr=subprocess.DEVNULL)
        else:
            subprocess.run(['rm', '-rf', path], check=True, stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError as e:
        if sudo_pwd and os.path.exists(path) and is_linux():
            run_command_sudo(['rm', '-rf', path], sudo_pwd)

    # Verify that the target is indeed deleted.
    if os.path.exists(path):
        print(f"Error: Failed to delete {path}")
        sys.exit(1)

def force_delete_glob(path: str, pattern: str, sudo_pwd: str = ""):
    # Will exit on any error.
    glob_targets = path_join(path, pattern)
    for target in glob.glob(glob_targets):
        force_delete(target, sudo_pwd)

def calculate_file_sha256(file_path: str) -> str:
    result = run_command(["shasum", "-a", "256", file_path])
    return result.split()[0]



import os

def log_params(host: str, package_file_path: str, temp_dir: str, version: str, sudo_pwd: str):
    print(f"Testing ta-lib-python installation on {host}")
    # Never display or log sudo_pwd, but want to know if it was specified.
    if sudo_pwd:
        hidden_sudo_pwd = "(hidden)"
    else:
        hidden_sudo_pwd = "\"\""

    print(f"  package_file_path={package_file_path}")
    print(f"  temp_dir={temp_dir}")
    print(f"  version={version}")
    print(f"  sudo_pwd={hidden_sudo_pwd}")

    # Create a dummy file "PARAMS" into temp_dir to help debugging.
    with open(os.path.join(temp_dir, "PARAMS"), "w") as f:
        f.write(f"package_file_path={package_file_path}\n")
        f.write(f"temp_dir={temp_dir}\n")
        f.write(f"version={version}\n")
        f.write(f"sudo_pwd={hidden_sudo_pwd}\n")

    return

def test_python_windows(package_file_path: str, temp_dir: str, version: str, sudo_pwd: str):
    # Test installation procedure for ta-lib-python and validate
    # that this ta-lib package release candidate is OK.
    #
    # Parameters
    # ----------
    # package_file_path is the ta-lib-{version}-win64.zip
    #
    # temp_dir is an empty directory that can be used
    # for most intermediate file operations.
    #
    # version is the "0.0.0" string expected to be returned
    # by the installed ta-lib package.
    #
    # sudo_pwd can be pipeline into a "sudo -S" command.
    # If sudo_pwd is an empty string, then call "sudo" without
    # "-S" and let it be interactive.
    # More info:
    #    https://stackoverflow.com/questions/60807449/run-github-action-as-sudo
    #
    # Test Behavior
    # -------------
    # Must re-install/upgrade if ta-lib is already installed.
    #
    # Recommended to create a venv in temp_dir to simulate
    # a user setup.
    #
    # Leave all intermediate files in temp_dir after the test
    # (might be useful for debugging).
    #
    # Just sys.exit(1) on any problem discovered.

    log_params("Windows", package_file_path, temp_dir, version, sudo_pwd)

    # TODO - Implement the test !!!

    return

def test_python_linux(package_file_path: str, temp_dir: str, version: str, sudo_pwd: str):
    # Same as test_python_windows except package_file_path is
    # the ta-lib-{version}-src.tar.gz
    #
    # Installation to be done with autotools "./configure"

    log_params("Linux", package_file_path, temp_dir, version, sudo_pwd)

    # TODO - Implement the test !!!

    return



#!/usr/bin/env python3

# Update *local* dev with latest from both dev and main *remote* branch.
#
# Such merging is often needed prior to a 'git push'.
#
# The local dev changes (if any) are temporarly stashed and merge back...
# so conflicts may need to be resolved manually (an error will be displayed).
#
# NOOP if nothing to merge or sync.
#
# Other processing done:
#  - Sync the TA-Lib versioning (See VERSION file) consistently with various other files.
#  - Update the TA-Lib source digest in ta_common.h (as needed).
#


from datetime import datetime
import random
import string
import subprocess
import sys

from utilities.common import verify_git_repo, run_command
from utilities.versions import sync_sources_digest, sync_versions

def generate_short_unique_id(length=20) -> str:
    # Generate a "unique enough" short identifier.
    timestamp = datetime.now().strftime('%Y%m%d%H%M%S')
    random_str = ''.join(random.choices(string.ascii_letters + string.digits, k=length - len(timestamp)))
    return timestamp + '-' + random_str

def main():
    try:
        original_branch = None

        # Switch to dev branch if not already on it
        root_dir = verify_git_repo()
        original_branch = run_command(['git', 'rev-parse', '--abbrev-ref', 'HEAD'])

        # Do nothing if there is staged changes.
        try:
            run_command(['git', 'diff', '--cached', '--exit-code'])
        except subprocess.CalledProcessError:
            print("Info: staged git changes detected. This script is intended to be run **before** any staging. No sync done.")
            sys.exit(1)

        # Do nothing if there are local commits not yet pushed.
        if run_command(['git', 'rev-list', '@{u}..HEAD']):
            print("Info: local commits need to be pushed. This script is intended to be run **before** any local staging/commits. No sync done.")
            sys.exit(1)

        # Fetch the latest branch information from origin
        run_command(['git', 'fetch', 'origin'])

        # Stash any local dev changes with a unique message
        stash_message = f'sync-script-stash-{generate_short_unique_id()}'

        if original_branch != "dev":
            print("Switching to dev branch")
            run_command(['git', 'checkout', 'dev'])
        run_command(['git', 'stash', 'push', '-m', stash_message])

        # Since everything in dev is now stashed, it is now possible
        # to switch to main branch and sync it.
        run_command(['git', 'checkout', 'main'])
        run_command(['git', 'pull', 'origin', 'main'])
        run_command(['git', 'checkout', 'dev'])

        # Get the local dev commit hash before pulling
        # This is later used to detect if any changes were pulled.
        local_dev_commit_before = run_command(['git', 'rev-parse', 'dev'])

        # Pull the latest dev changes
        run_command(['git', 'pull', 'origin', 'dev'])

        # Check if there were any changes pulled
        local_dev_commit_after = run_command(['git', 'rev-parse', 'dev'])
        if local_dev_commit_before == local_dev_commit_after:
            print("No changes to merge from origin/dev")
        else:
            print("Pulled latest changes from origin/dev")

        # Apply the stashed changes
        stash_list = run_command(['git', 'stash', 'list'])
        if stash_message in stash_list:
            try:
                run_command(['git', 'stash', 'pop'])
            except subprocess.CalledProcessError:
                print("Conflict occurred while applying stashed changes. Resolve conflicts manually.")
                print("1. Identify conflicts with 'git status'")
                print("2. Resolve manually by editing the conflicted files")
                print("3. Mark as resolved using 'git add <file>'")
                print("4. Complete merge with a 'git commit'")
                sys.exit(1)


        # Switch back to dev branch
        run_command(['git', 'checkout', 'dev'])

        # Find the common ancestor of dev and main
        merge_base = run_command(['git', 'merge-base', 'dev', 'main'])

        # Check if there are any changes from main that are not in dev
        diff_output = subprocess.run(['git', 'diff', '--quiet', merge_base, 'main'], stderr=subprocess.DEVNULL)
        if diff_output.returncode == 0:
            print("No changes to merge from origin/main")
        else:
            # Perform the actual merge
            merge_output = subprocess.run(['git', 'merge', '--no-commit', '--no-ff', 'main'], stderr=subprocess.DEVNULL)
            if merge_output.returncode == 0:
                # Check if there are any changes to commit
                diff_index_output = subprocess.run(['git', 'diff-index', '--quiet', 'HEAD', '--'], stderr=subprocess.DEVNULL)
                if diff_index_output.returncode == 0:
                    print("No changes to merge from main to dev.")
                    run_command(['git', 'merge', '--abort'])
                else:
                    # Commit the merge if there are changes
                    run_command(['git', 'commit', '-m', 'Merged main into dev'])
                    print("Merged main into dev.")
            else:
                print("Merge failed due to conflicts. Next steps:")
                print("1. Identify conflicts with 'git status'")
                print("2. Resolve manually by editing the conflicted files")
                print("3. Mark as resolved using 'git add <file>'")
                print("4. Complete merge with a 'git commit'")
                sys.exit(1)

        # Make sure TA-Lib versioning is consistent in various files
        # used for building packages.
        is_updated, version = sync_versions(root_dir)
        if not is_updated:
            print(f"No changes to version [{version}]")

        # Update TA_LIB_SOURCES_DIGEST in ta_common.h (as needed)
        is_updated, digest = sync_sources_digest(root_dir)
        if is_updated:
            print(f"Updated sources digest (ta_common.h): [{digest}]")
        else:
            print(f"No changes to sources digest (ta_common.h) [{digest}]")

    except subprocess.CalledProcessError as e:
        print(f"An error occurred: {e}")
        sys.exit(1)

    finally:
        # Restore to the branch the user was located before running this script
        if original_branch:
            current_branch = run_command(['git', 'rev-parse', '--abbrev-ref', 'HEAD'])
            if current_branch != original_branch:
                print(f"Switching back to {original_branch} branch")
                run_command(['git', 'checkout', original_branch])

if __name__ == "__main__":
    main()



import numpy as np
import atexit

# We import the compiled extension module
from . import pytafast_ext
from .pytafast_ext import MAType

__version__ = "0.2.0"

def _is_pandas_series(obj):
    try:
        import pandas as pd
        return isinstance(obj, pd.Series)
    except ImportError:
        return False

def SMA(inReal, timeperiod=30):
    """
    Simple Moving Average.
    
    Args:
        inReal: 1D array-like (numpy ndarray, pandas Series, list, etc.) of floats.
        timeperiod: Integer time period.
        
    Returns:
        A numpy array or pandas Series (matching input) with the SMA values.
        The leading (timeperiod - 1) elements will be NaN.
    """
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.SMA(arr, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name=inReal.name)
    return out_arr

def EMA(inReal, timeperiod=30):
    """
    Exponential Moving Average.
    """
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.EMA(arr, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name=inReal.name)
    return out_arr

def RSI(inReal, timeperiod=14):
    """
    Relative Strength Index.
    """
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.RSI(arr, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name=inReal.name)
    return out_arr

def MACD(inReal, fastperiod=12, slowperiod=26, signalperiod=9):
    """
    Moving Average Convergence/Divergence.
    Returns: (macd, macdsignal, macdhist)
    """
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    macd, macdsignal, macdhist = pytafast_ext.MACD(arr, fastperiod, slowperiod, signalperiod)
    if is_series:
        import pandas as pd
        return (
            pd.Series(macd, index=inReal.index, name="MACD"),
            pd.Series(macdsignal, index=inReal.index, name="MACD_Signal"),
            pd.Series(macdhist, index=inReal.index, name="MACD_Hist")
        )
    return macd, macdsignal, macdhist

def BBANDS(inReal, timeperiod=5, nbdevup=2.0, nbdevdn=2.0, matype=MAType.SMA):
    """
    Bollinger Bands.
    Returns: (upperband, middleband, lowerband)
    """
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    upper, middle, lower = pytafast_ext.BBANDS(arr, timeperiod, nbdevup, nbdevdn, int(matype.value) if hasattr(matype, 'value') else int(matype))
    
    if is_series:
        import pandas as pd
        return (
            pd.Series(upper, index=inReal.index, name="UpperBand"),
            pd.Series(middle, index=inReal.index, name="MiddleBand"),
            pd.Series(lower, index=inReal.index, name="LowerBand")
        )
    return upper, middle, lower

def ATR(inHigh, inLow, inClose, timeperiod=14):
    """
    Average True Range.
    """
    is_series = _is_pandas_series(inClose)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    arr_c = np.ascontiguousarray(inClose, dtype=np.float64)
    out_arr = pytafast_ext.ATR(arr_h, arr_l, arr_c, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inClose.index, name="ATR")
    return out_arr

def ADX(inHigh, inLow, inClose, timeperiod=14):
    """
    Average Directional Movement Index.
    """
    is_series = _is_pandas_series(inClose)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    arr_c = np.ascontiguousarray(inClose, dtype=np.float64)
    out_arr = pytafast_ext.ADX(arr_h, arr_l, arr_c, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inClose.index, name="ADX")
    return out_arr

def CCI(inHigh, inLow, inClose, timeperiod=14):
    """
    Commodity Channel Index.
    """
    is_series = _is_pandas_series(inClose)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    arr_c = np.ascontiguousarray(inClose, dtype=np.float64)
    out_arr = pytafast_ext.CCI(arr_h, arr_l, arr_c, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inClose.index, name="CCI")
    return out_arr

def OBV(inReal, inVolume):
    """
    On Balance Volume.
    """
    is_series = _is_pandas_series(inReal)
    arr_c = np.ascontiguousarray(inReal, dtype=np.float64)
    arr_v = np.ascontiguousarray(inVolume, dtype=np.float64)
    out_arr = pytafast_ext.OBV(arr_c, arr_v)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="OBV")
    return out_arr

def ROC(inReal, timeperiod=10):
    """
    Rate of change : ((price/prevPrice)-1)*100
    """
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.ROC(arr, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="ROC")
    return out_arr

def STOCH(inHigh, inLow, inClose, fastk_period=5, slowk_period=3, slowk_matype=MAType.SMA, slowd_period=3, slowd_matype=MAType.SMA):
    """
    Stochastic.
    Returns: (slowk, slowd)
    """
    is_series = _is_pandas_series(inClose)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    arr_c = np.ascontiguousarray(inClose, dtype=np.float64)
    
    sk_t = int(slowk_matype.value) if hasattr(slowk_matype, 'value') else int(slowk_matype)
    sd_t = int(slowd_matype.value) if hasattr(slowd_matype, 'value') else int(slowd_matype)
    
    slowk, slowd = pytafast_ext.STOCH(arr_h, arr_l, arr_c, fastk_period, slowk_period, sk_t, slowd_period, sd_t)
    
    if is_series:
        import pandas as pd
        return (
            pd.Series(slowk, index=inClose.index, name="SlowK"),
            pd.Series(slowd, index=inClose.index, name="SlowD")
        )
    return slowk, slowd

def MOM(inReal, timeperiod=10):
    """
    Momentum.
    """
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.MOM(arr, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="MOM")
    return out_arr

def STDDEV(inReal, timeperiod=5, nbdev=1.0):
    """
    Standard Deviation.
    """
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.STDDEV(arr, timeperiod, nbdev)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="STDDEV")
    return out_arr

def WILLR(inHigh, inLow, inClose, timeperiod=14):
    """
    Williams %R.
    """
    is_series = _is_pandas_series(inClose)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    arr_c = np.ascontiguousarray(inClose, dtype=np.float64)
    out_arr = pytafast_ext.WILLR(arr_h, arr_l, arr_c, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inClose.index, name="WILLR")
    return out_arr

def NATR(inHigh, inLow, inClose, timeperiod=14):
    """
    Normalized Average True Range.
    """
    is_series = _is_pandas_series(inClose)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    arr_c = np.ascontiguousarray(inClose, dtype=np.float64)
    out_arr = pytafast_ext.NATR(arr_h, arr_l, arr_c, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inClose.index, name="NATR")
    return out_arr

def MFI(inHigh, inLow, inClose, inVolume, timeperiod=14):
    """
    Money Flow Index.
    """
    is_series = _is_pandas_series(inClose)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    arr_c = np.ascontiguousarray(inClose, dtype=np.float64)
    arr_v = np.ascontiguousarray(inVolume, dtype=np.float64)
    out_arr = pytafast_ext.MFI(arr_h, arr_l, arr_c, arr_v, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inClose.index, name="MFI")
    return out_arr

def CMO(inReal, timeperiod=14):
    """
    Chande Momentum Oscillator.
    """
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.CMO(arr, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="CMO")
    return out_arr

def DX(inHigh, inLow, inClose, timeperiod=14):
    """
    Directional Movement Index.
    """
    is_series = _is_pandas_series(inClose)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    arr_c = np.ascontiguousarray(inClose, dtype=np.float64)
    out_arr = pytafast_ext.DX(arr_h, arr_l, arr_c, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inClose.index, name="DX")
    return out_arr

def MINUS_DI(inHigh, inLow, inClose, timeperiod=14):
    """
    Minus Directional Indicator.
    """
    is_series = _is_pandas_series(inClose)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    arr_c = np.ascontiguousarray(inClose, dtype=np.float64)
    out_arr = pytafast_ext.MINUS_DI(arr_h, arr_l, arr_c, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inClose.index, name="MINUS_DI")
    return out_arr

def MINUS_DM(inHigh, inLow, timeperiod=14):
    """
    Minus Directional Movement.
    """
    is_series = _is_pandas_series(inHigh)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    out_arr = pytafast_ext.MINUS_DM(arr_h, arr_l, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inHigh.index, name="MINUS_DM")
    return out_arr

def PLUS_DI(inHigh, inLow, inClose, timeperiod=14):
    """
    Plus Directional Indicator.
    """
    is_series = _is_pandas_series(inClose)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    arr_c = np.ascontiguousarray(inClose, dtype=np.float64)
    out_arr = pytafast_ext.PLUS_DI(arr_h, arr_l, arr_c, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inClose.index, name="PLUS_DI")
    return out_arr

def PLUS_DM(inHigh, inLow, timeperiod=14):
    """
    Plus Directional Movement.
    """
    is_series = _is_pandas_series(inHigh)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    out_arr = pytafast_ext.PLUS_DM(arr_h, arr_l, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inHigh.index, name="PLUS_DM")
    return out_arr

def APO(inReal, fastperiod=12, slowperiod=26, matype=0):
    """
    Absolute Price Oscillator.
    """
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.APO(arr, fastperiod, slowperiod, matype)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="APO")
    return out_arr

def AROON(inHigh, inLow, timeperiod=14):
    """
    Aroon.
    """
    is_series = _is_pandas_series(inHigh)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    out_down, out_up = pytafast_ext.AROON(arr_h, arr_l, timeperiod)
    if is_series:
        import pandas as pd
        return (
            pd.Series(out_down, index=inHigh.index, name="AROON_DOWN"),
            pd.Series(out_up, index=inHigh.index, name="AROON_UP")
        )
    return out_down, out_up

def AROONOSC(inHigh, inLow, timeperiod=14):
    """
    Aroon Oscillator.
    """
    is_series = _is_pandas_series(inHigh)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    out_arr = pytafast_ext.AROONOSC(arr_h, arr_l, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inHigh.index, name="AROONOSC")
    return out_arr

def PPO(inReal, fastperiod=12, slowperiod=26, matype=0):
    """
    Percentage Price Oscillator.
    """
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.PPO(arr, fastperiod, slowperiod, matype)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="PPO")
    return out_arr

def TRIX(inReal, timeperiod=30):
    """
    1-day Rate-Of-Change (ROC) of a Triple Smooth EMA.
    """
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.TRIX(arr, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="TRIX")
    return out_arr

def ULTOSC(inHigh, inLow, inClose, timeperiod1=7, timeperiod2=14, timeperiod3=28):
    """
    Ultimate Oscillator.
    """
    is_series = _is_pandas_series(inClose)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    arr_c = np.ascontiguousarray(inClose, dtype=np.float64)
    out_arr = pytafast_ext.ULTOSC(arr_h, arr_l, arr_c, timeperiod1, timeperiod2, timeperiod3)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inClose.index, name="ULTOSC")
    return out_arr

def DEMA(inReal, timeperiod=30):
    """Double Exponential Moving Average."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.DEMA(arr, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="DEMA")
    return out_arr

def KAMA(inReal, timeperiod=30):
    """Kaufman Adaptive Moving Average."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.KAMA(arr, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="KAMA")
    return out_arr

def MA(inReal, timeperiod=30, matype=0):
    """Moving Average (generic)."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.MA(arr, timeperiod, matype)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="MA")
    return out_arr

def T3(inReal, timeperiod=5, vfactor=0.7):
    """Triple Exponential Moving Average (T3)."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.T3(arr, timeperiod, vfactor)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="T3")
    return out_arr

def TEMA(inReal, timeperiod=30):
    """Triple Exponential Moving Average."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.TEMA(arr, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="TEMA")
    return out_arr

def TRIMA(inReal, timeperiod=30):
    """Triangular Moving Average."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.TRIMA(arr, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="TRIMA")
    return out_arr

def WMA(inReal, timeperiod=30):
    """Weighted Moving Average."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.WMA(arr, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="WMA")
    return out_arr

def SAR(inHigh, inLow, acceleration=0.02, maximum=0.2):
    """Parabolic SAR."""
    is_series = _is_pandas_series(inHigh)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    out_arr = pytafast_ext.SAR(arr_h, arr_l, acceleration, maximum)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inHigh.index, name="SAR")
    return out_arr

def TRANGE(inHigh, inLow, inClose):
    """True Range."""
    is_series = _is_pandas_series(inClose)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    arr_c = np.ascontiguousarray(inClose, dtype=np.float64)
    out_arr = pytafast_ext.TRANGE(arr_h, arr_l, arr_c)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inClose.index, name="TRANGE")
    return out_arr

def AVGPRICE(inOpen, inHigh, inLow, inClose):
    """Average Price."""
    is_series = _is_pandas_series(inClose)
    arr_o = np.ascontiguousarray(inOpen, dtype=np.float64)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    arr_c = np.ascontiguousarray(inClose, dtype=np.float64)
    out_arr = pytafast_ext.AVGPRICE(arr_o, arr_h, arr_l, arr_c)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inClose.index, name="AVGPRICE")
    return out_arr

def MEDPRICE(inHigh, inLow):
    """Median Price."""
    is_series = _is_pandas_series(inHigh)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    out_arr = pytafast_ext.MEDPRICE(arr_h, arr_l)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inHigh.index, name="MEDPRICE")
    return out_arr

def TYPPRICE(inHigh, inLow, inClose):
    """Typical Price."""
    is_series = _is_pandas_series(inClose)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    arr_c = np.ascontiguousarray(inClose, dtype=np.float64)
    out_arr = pytafast_ext.TYPPRICE(arr_h, arr_l, arr_c)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inClose.index, name="TYPPRICE")
    return out_arr

def WCLPRICE(inHigh, inLow, inClose):
    """Weighted Close Price."""
    is_series = _is_pandas_series(inClose)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    arr_c = np.ascontiguousarray(inClose, dtype=np.float64)
    out_arr = pytafast_ext.WCLPRICE(arr_h, arr_l, arr_c)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inClose.index, name="WCLPRICE")
    return out_arr

def MIDPOINT(inReal, timeperiod=14):
    """MidPoint over period."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.MIDPOINT(arr, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="MIDPOINT")
    return out_arr

def MIDPRICE(inHigh, inLow, timeperiod=14):
    """Midpoint Price over period."""
    is_series = _is_pandas_series(inHigh)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    out_arr = pytafast_ext.MIDPRICE(arr_h, arr_l, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inHigh.index, name="MIDPRICE")
    return out_arr

# Initialize TA-Lib context when the module is loaded
pytafast_ext.initialize()
atexit.register(pytafast_ext.shutdown)

def ADXR(inHigh, inLow, inClose, timeperiod=14):
    """Average Directional Movement Index Rating."""
    is_series = _is_pandas_series(inClose)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    arr_c = np.ascontiguousarray(inClose, dtype=np.float64)
    out_arr = pytafast_ext.ADXR(arr_h, arr_l, arr_c, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inClose.index, name="ADXR")
    return out_arr

def BOP(inOpen, inHigh, inLow, inClose):
    """Balance Of Power."""
    is_series = _is_pandas_series(inClose)
    arr_o = np.ascontiguousarray(inOpen, dtype=np.float64)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    arr_c = np.ascontiguousarray(inClose, dtype=np.float64)
    out_arr = pytafast_ext.BOP(arr_o, arr_h, arr_l, arr_c)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inClose.index, name="BOP")
    return out_arr

def MACDEXT(inReal, fastperiod=12, fastmatype=0, slowperiod=26, slowmatype=0,
            signalperiod=9, signalmatype=0):
    """MACD with controllable MA type."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    macd, signal, hist = pytafast_ext.MACDEXT(arr, fastperiod, fastmatype,
                                               slowperiod, slowmatype,
                                               signalperiod, signalmatype)
    if is_series:
        import pandas as pd
        idx = inReal.index
        return (pd.Series(macd, index=idx, name="MACD"),
                pd.Series(signal, index=idx, name="MACDSignal"),
                pd.Series(hist, index=idx, name="MACDHist"))
    return macd, signal, hist

def MACDFIX(inReal, signalperiod=9):
    """MACD Fix 12/26."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    macd, signal, hist = pytafast_ext.MACDFIX(arr, signalperiod)
    if is_series:
        import pandas as pd
        idx = inReal.index
        return (pd.Series(macd, index=idx, name="MACD"),
                pd.Series(signal, index=idx, name="MACDSignal"),
                pd.Series(hist, index=idx, name="MACDHist"))
    return macd, signal, hist

def STOCHF(inHigh, inLow, inClose, fastk_period=5, fastd_period=3,
           fastd_matype=0):
    """Stochastic Fast."""
    is_series = _is_pandas_series(inClose)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    arr_c = np.ascontiguousarray(inClose, dtype=np.float64)
    fastk, fastd = pytafast_ext.STOCHF(arr_h, arr_l, arr_c, fastk_period,
                                        fastd_period, fastd_matype)
    if is_series:
        import pandas as pd
        idx = inClose.index
        return (pd.Series(fastk, index=idx, name="FastK"),
                pd.Series(fastd, index=idx, name="FastD"))
    return fastk, fastd

def STOCHRSI(inReal, timeperiod=14, fastk_period=5, fastd_period=3,
             fastd_matype=0):
    """Stochastic RSI."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    fastk, fastd = pytafast_ext.STOCHRSI(arr, timeperiod, fastk_period,
                                          fastd_period, fastd_matype)
    if is_series:
        import pandas as pd
        idx = inReal.index
        return (pd.Series(fastk, index=idx, name="FastK"),
                pd.Series(fastd, index=idx, name="FastD"))
    return fastk, fastd

def ROCP(inReal, timeperiod=10):
    """Rate of change Percentage."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.ROCP(arr, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="ROCP")
    return out_arr

def ROCR(inReal, timeperiod=10):
    """Rate of change ratio."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.ROCR(arr, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="ROCR")
    return out_arr

def ROCR100(inReal, timeperiod=10):
    """Rate of change ratio 100 scale."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.ROCR100(arr, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="ROCR100")
    return out_arr

def AD(inHigh, inLow, inClose, inVolume):
    """Chaikin A/D Line."""
    is_series = _is_pandas_series(inClose)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    arr_c = np.ascontiguousarray(inClose, dtype=np.float64)
    arr_v = np.ascontiguousarray(inVolume, dtype=np.float64)
    out_arr = pytafast_ext.AD(arr_h, arr_l, arr_c, arr_v)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inClose.index, name="AD")
    return out_arr

def ADOSC(inHigh, inLow, inClose, inVolume, fastperiod=3, slowperiod=10):
    """Chaikin A/D Oscillator."""
    is_series = _is_pandas_series(inClose)
    arr_h = np.ascontiguousarray(inHigh, dtype=np.float64)
    arr_l = np.ascontiguousarray(inLow, dtype=np.float64)
    arr_c = np.ascontiguousarray(inClose, dtype=np.float64)
    arr_v = np.ascontiguousarray(inVolume, dtype=np.float64)
    out_arr = pytafast_ext.ADOSC(arr_h, arr_l, arr_c, arr_v, fastperiod, slowperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inClose.index, name="ADOSC")
    return out_arr

def BETA(inReal0, inReal1, timeperiod=5):
    """Beta."""
    is_series = _is_pandas_series(inReal0)
    arr0 = np.ascontiguousarray(inReal0, dtype=np.float64)
    arr1 = np.ascontiguousarray(inReal1, dtype=np.float64)
    out_arr = pytafast_ext.BETA(arr0, arr1, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal0.index, name="BETA")
    return out_arr

def CORREL(inReal0, inReal1, timeperiod=30):
    """Pearson's Correlation Coefficient."""
    is_series = _is_pandas_series(inReal0)
    arr0 = np.ascontiguousarray(inReal0, dtype=np.float64)
    arr1 = np.ascontiguousarray(inReal1, dtype=np.float64)
    out_arr = pytafast_ext.CORREL(arr0, arr1, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal0.index, name="CORREL")
    return out_arr

def LINEARREG(inReal, timeperiod=14):
    """Linear Regression."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.LINEARREG(arr, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="LINEARREG")
    return out_arr

def LINEARREG_ANGLE(inReal, timeperiod=14):
    """Linear Regression Angle."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.LINEARREG_ANGLE(arr, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="LINEARREG_ANGLE")
    return out_arr

def LINEARREG_INTERCEPT(inReal, timeperiod=14):
    """Linear Regression Intercept."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.LINEARREG_INTERCEPT(arr, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="LINEARREG_INTERCEPT")
    return out_arr

def LINEARREG_SLOPE(inReal, timeperiod=14):
    """Linear Regression Slope."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.LINEARREG_SLOPE(arr, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="LINEARREG_SLOPE")
    return out_arr

def TSF(inReal, timeperiod=14):
    """Time Series Forecast."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.TSF(arr, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="TSF")
    return out_arr

def VAR(inReal, timeperiod=5, nbdev=1.0):
    """Variance."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.VAR(arr, timeperiod, nbdev)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="VAR")
    return out_arr

def AVGDEV(inReal, timeperiod=14):
    """Average Deviation."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.AVGDEV(arr, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="AVGDEV")
    return out_arr

def ADD(inReal0, inReal1):
    """Vector Arithmetic Add."""
    is_series = _is_pandas_series(inReal0)
    arr0 = np.ascontiguousarray(inReal0, dtype=np.float64)
    arr1 = np.ascontiguousarray(inReal1, dtype=np.float64)
    out_arr = pytafast_ext.ADD(arr0, arr1)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal0.index, name="ADD")
    return out_arr

def SUB(inReal0, inReal1):
    """Vector Arithmetic Subtraction."""
    is_series = _is_pandas_series(inReal0)
    arr0 = np.ascontiguousarray(inReal0, dtype=np.float64)
    arr1 = np.ascontiguousarray(inReal1, dtype=np.float64)
    out_arr = pytafast_ext.SUB(arr0, arr1)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal0.index, name="SUB")
    return out_arr

def MULT(inReal0, inReal1):
    """Vector Arithmetic Multiply."""
    is_series = _is_pandas_series(inReal0)
    arr0 = np.ascontiguousarray(inReal0, dtype=np.float64)
    arr1 = np.ascontiguousarray(inReal1, dtype=np.float64)
    out_arr = pytafast_ext.MULT(arr0, arr1)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal0.index, name="MULT")
    return out_arr

def DIV(inReal0, inReal1):
    """Vector Arithmetic Division."""
    is_series = _is_pandas_series(inReal0)
    arr0 = np.ascontiguousarray(inReal0, dtype=np.float64)
    arr1 = np.ascontiguousarray(inReal1, dtype=np.float64)
    out_arr = pytafast_ext.DIV(arr0, arr1)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal0.index, name="DIV")
    return out_arr

def _make_math_transform(name):
    """Factory for single-input math transform wrappers."""
    def wrapper(inReal):
        is_series = _is_pandas_series(inReal)
        arr = np.ascontiguousarray(inReal, dtype=np.float64)
        out_arr = getattr(pytafast_ext, name)(arr)
        if is_series:
            import pandas as pd
            return pd.Series(out_arr, index=inReal.index, name=name)
        return out_arr
    wrapper.__name__ = name
    wrapper.__doc__ = f"Vector {name}."
    return wrapper

ACOS = _make_math_transform("ACOS")
ASIN = _make_math_transform("ASIN")
ATAN = _make_math_transform("ATAN")
CEIL = _make_math_transform("CEIL")
COS = _make_math_transform("COS")
COSH = _make_math_transform("COSH")
EXP = _make_math_transform("EXP")
FLOOR = _make_math_transform("FLOOR")
LN = _make_math_transform("LN")
LOG10 = _make_math_transform("LOG10")
SIN = _make_math_transform("SIN")
SINH = _make_math_transform("SINH")
SQRT = _make_math_transform("SQRT")
TAN = _make_math_transform("TAN")
TANH = _make_math_transform("TANH")

def MAX(inReal, timeperiod=30):
    """Highest value over a specified period."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.MAX(arr, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="MAX")
    return out_arr

def MIN(inReal, timeperiod=30):
    """Lowest value over a specified period."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.MIN(arr, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="MIN")
    return out_arr

def SUM(inReal, timeperiod=30):
    """Summation."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.SUM(arr, timeperiod)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="SUM")
    return out_arr

def MINMAX(inReal, timeperiod=30):
    """Lowest and highest values over a specified period."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_min, out_max = pytafast_ext.MINMAX(arr, timeperiod)
    if is_series:
        import pandas as pd
        return (pd.Series(out_min, index=inReal.index, name="min"),
                pd.Series(out_max, index=inReal.index, name="max"))
    return out_min, out_max

def MINMAXINDEX(inReal, timeperiod=30):
    """Indexes of lowest and highest values over a specified period."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_minidx, out_maxidx = pytafast_ext.MINMAXINDEX(arr, timeperiod)
    if is_series:
        import pandas as pd
        return (pd.Series(out_minidx, index=inReal.index, name="minidx"),
                pd.Series(out_maxidx, index=inReal.index, name="maxidx"))
    return out_minidx, out_maxidx

def HT_DCPERIOD(inReal):
    """Hilbert Transform - Dominant Cycle Period."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.HT_DCPERIOD(arr)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="HT_DCPERIOD")
    return out_arr

def HT_DCPHASE(inReal):
    """Hilbert Transform - Dominant Cycle Phase."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.HT_DCPHASE(arr)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="HT_DCPHASE")
    return out_arr

def HT_PHASOR(inReal):
    """Hilbert Transform - Phasor Components."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    inphase, quadrature = pytafast_ext.HT_PHASOR(arr)
    if is_series:
        import pandas as pd
        return (pd.Series(inphase, index=inReal.index, name="inphase"),
                pd.Series(quadrature, index=inReal.index, name="quadrature"))
    return inphase, quadrature

def HT_SINE(inReal):
    """Hilbert Transform - SineWave."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    sine, leadsine = pytafast_ext.HT_SINE(arr)
    if is_series:
        import pandas as pd
        return (pd.Series(sine, index=inReal.index, name="sine"),
                pd.Series(leadsine, index=inReal.index, name="leadsine"))
    return sine, leadsine

def HT_TRENDLINE(inReal):
    """Hilbert Transform - Instantaneous Trendline."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.HT_TRENDLINE(arr)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="HT_TRENDLINE")
    return out_arr

def HT_TRENDMODE(inReal):
    """Hilbert Transform - Trend vs Cycle Mode."""
    is_series = _is_pandas_series(inReal)
    arr = np.ascontiguousarray(inReal, dtype=np.float64)
    out_arr = pytafast_ext.HT_TRENDMODE(arr)
    if is_series:
        import pandas as pd
        return pd.Series(out_arr, index=inReal.index, name="HT_TRENDMODE")
    return out_arr

# --- Candlestick Patterns ---
_CDL_STANDARD = [
    "CDL2CROWS", "CDL3BLACKCROWS", "CDL3INSIDE", "CDL3LINESTRIKE",
    "CDL3OUTSIDE", "CDL3STARSINSOUTH", "CDL3WHITESOLDIERS",
    "CDLADVANCEBLOCK", "CDLBELTHOLD", "CDLBREAKAWAY",
    "CDLCLOSINGMARUBOZU", "CDLCONCEALBABYSWALL", "CDLCOUNTERATTACK",
    "CDLDOJI", "CDLDOJISTAR", "CDLDRAGONFLYDOJI", "CDLENGULFING",
    "CDLGAPSIDESIDEWHITE", "CDLGRAVESTONEDOJI", "CDLHAMMER",
    "CDLHANGINGMAN", "CDLHARAMI", "CDLHARAMICROSS", "CDLHIGHWAVE",
    "CDLHIKKAKE", "CDLHIKKAKEMOD", "CDLHOMINGPIGEON",
    "CDLIDENTICAL3CROWS", "CDLINNECK", "CDLINVERTEDHAMMER",
    "CDLKICKING", "CDLKICKINGBYLENGTH", "CDLLADDERBOTTOM",
    "CDLLONGLEGGEDDOJI", "CDLLONGLINE", "CDLMARUBOZU",
    "CDLMATCHINGLOW", "CDLONNECK", "CDLPIERCING", "CDLRICKSHAWMAN",
    "CDLRISEFALL3METHODS", "CDLSEPARATINGLINES", "CDLSHOOTINGSTAR",
    "CDLSHORTLINE", "CDLSPINNINGTOP", "CDLSTALLEDPATTERN",
    "CDLSTICKSANDWICH", "CDLTAKURI", "CDLTASUKIGAP", "CDLTHRUSTING",
    "CDLTRISTAR", "CDLUNIQUE3RIVER", "CDLUPSIDEGAP2CROWS",
    "CDLXSIDEGAP3METHODS",
]

_CDL_PENETRATION = {
    "CDLABANDONEDBABY": 0.3, "CDLDARKCLOUDCOVER": 0.5,
    "CDLEVENINGDOJISTAR": 0.3, "CDLEVENINGSTAR": 0.3,
    "CDLMATHOLD": 0.5, "CDLMORNINGDOJISTAR": 0.3, "CDLMORNINGSTAR": 0.3,
}

def _make_cdl_standard(name):
    def wrapper(inOpen, inHigh, inLow, inClose):
        is_series = _is_pandas_series(inClose)
        o = np.ascontiguousarray(inOpen, dtype=np.float64)
        h = np.ascontiguousarray(inHigh, dtype=np.float64)
        l = np.ascontiguousarray(inLow, dtype=np.float64)
        c = np.ascontiguousarray(inClose, dtype=np.float64)
        out = getattr(pytafast_ext, name)(o, h, l, c)
        if is_series:
            import pandas as pd
            return pd.Series(out, index=inClose.index, name=name)
        return out
    wrapper.__name__ = name
    wrapper.__doc__ = f"Candlestick Pattern: {name}"
    return wrapper

def _make_cdl_penetration(name, default_pen):
    def wrapper(inOpen, inHigh, inLow, inClose, penetration=default_pen):
        is_series = _is_pandas_series(inClose)
        o = np.ascontiguousarray(inOpen, dtype=np.float64)
        h = np.ascontiguousarray(inHigh, dtype=np.float64)
        l = np.ascontiguousarray(inLow, dtype=np.float64)
        c = np.ascontiguousarray(inClose, dtype=np.float64)
        out = getattr(pytafast_ext, name)(o, h, l, c, penetration)
        if is_series:
            import pandas as pd
            return pd.Series(out, index=inClose.index, name=name)
        return out
    wrapper.__name__ = name
    wrapper.__doc__ = f"Candlestick Pattern: {name}"
    return wrapper

# Generate all standard CDL wrappers
for _name in _CDL_STANDARD:
    globals()[_name] = _make_cdl_standard(_name)

# Generate all penetration CDL wrappers
for _name, _pen in _CDL_PENETRATION.items():
    globals()[_name] = _make_cdl_penetration(_name, _pen)

from pytafast import aio






import os
import sys
import tempfile
from utilities.common import run_command
from utilities.files import path_join

vcvarsall_env_created = False

def call_vcvarsall(root_dir: str, args: list):
    global vcvarsall_env_created

    # If first time, then create a temp/vcvarsall_env.txt file
    # with all the current env variables.
    #
    # If this function found that temp/vcvarsall_env.txt exsists, then
    # it will restore the env from that file before calling vcvarsall.bat
    #
    # This is needed to avoid PATH pollution with each subsequent call 
    # to vcvarsall.bat
    temp_dir = path_join(root_dir, "temp")
    os.makedirs(temp_dir, exist_ok=True)
    vcvarsall_env_file = path_join(temp_dir, "vcvarsall_env.txt")
    try:
        if not vcvarsall_env_created:
            # Delete the file if it exists
            if os.path.exists(vcvarsall_env_file):
                os.remove(vcvarsall_env_file)
            with open(vcvarsall_env_file, 'w') as f:
                for key, value in os.environ.items():
                    f.write(f'{key}={value}\n')
            # Verify that vcvarsall_env_file exists.
            if os.path.exists(vcvarsall_env_file):
                vcvarsall_env_created = True
            else:
                print(f"Error: Could not create vcvarsall: {vcvarsall_env_file}")
                sys.exit(1)
        else:
            if not os.path.exists(vcvarsall_env_file):
                print(f"Error: vcvarsall_env.txt was not created at {vcvarsall_env_file}")
                sys.exit(1)

            with open(vcvarsall_env_file, 'r') as f:
                for line in f:
                    if line.strip():
                        key, value = line.strip().split('=', 1)
                        os.environ[key] = value
    except Exception as e:
        print(f"Error: {e}")
        sys.exit(1)

    if not sys.platform == "win32":
        print(f"Unexpected call to vcvarsall. Not supported on this platform")
        sys.exit(1)
    
    vswhere_path = r"C:\Program Files (x86)\Microsoft Visual Studio\Installer\vswhere.exe"
    if not os.path.exists(vswhere_path):
        print(f"Error: vswhere.exe not found at {vswhere_path}")
        sys.exit(1)

    # Print out the PATH variable
    #print(f"PATH before vcvarsall: {os.environ['PATH']}")

    vswhere_command = [vswhere_path, "-latest", "-products", "*", "-requires", "Microsoft.Component.MSBuild", "-find", "VC/Auxiliary/Build/vcvarsall.bat"]
    vcvarsall_path = run_command(vswhere_command)
    if not os.path.exists(vcvarsall_path):
        print(f"Error: vcvarsall.bat not found at {vcvarsall_path}")
        sys.exit(1)
    # Create a temporary batch file to call vcvarsall.bat and save a copy
    # of the whole env, so we can re-inject these into this python process.
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_batch_file = path_join(temp_dir, "temp_vcvarsall.bat")
        temp_env_file = path_join(temp_dir, "temp_env.txt")
        try:
            with open(temp_batch_file, 'w') as f:
                args_str = ' '.join(args)
                f.write(f'call "{vcvarsall_path}" {args_str}\n')
                f.write(f'set > "{temp_env_file}"\n')
            # Execute the temporary batch file, reinject the env.
            run_command([temp_batch_file])    
            with open(temp_env_file, 'r') as f:
                for line in f:
                    if line.strip():
                        key, value = line.strip().split('=', 1)
                        os.environ[key] = value

            # Print out the PATH variable
            #print(f"PATH after vcvarsall: {os.environ['PATH']}")

        except Exception as e:
            print(f"Error: {e}")
            sys.exit(1)

    # Verify that platform is set.
    if not os.environ.get('Platform', None):
        print(f"Error: Platform not set by vcvarsall.bat")
        sys.exit(1)
    print(f"vcvarsall.bat called successfully")
    print(f"Platform: {os.environ['Platform']}")



from dataclasses import dataclass
import hashlib
import json
import os
import sys

from utilities.files import path_join

def _calculate_md5(filepath: str) -> str:
    # Calculate md5 of a binary file (do not use for text file because
    # of portability issue for line endings)
    hash_md5 = hashlib.md5()
    with open(filepath, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()

def _digests_path(root_dir: str) -> str:
    return path_join(root_dir, "dist", "digests")

def _asset_file_name_to_filepath(root_dir: str, asset_file_name: str) -> str:
    return path_join(_digests_path(root_dir), f"{asset_file_name}.digest")

@dataclass
class PackageDigest:
    root_dir: str = ""          # Mandatory at construction
    asset_file_name: str = ""        # Mandatory at construction
    sources_digest: str = ""    # Mandatory at construction
    builder_id: str = ""        # Mandatory at construction

    built_success: str = "False"    # "True" or "False"

    package_md5: str = "Disabled"     # "Disabled", "Unknown" or "hash_of_package"
    gen_code_pass: str = "Disabled"   # "Disabled", "Unknown", "True" or "False"
    ta_regtest_pass: str = "Disabled" # "Disabled", "Unknown", "True" or "False"
    dist_test_pass: str = "Disabled"  # "Disabled", ""Unknown", True" or "False"

    @staticmethod
    def default(root_dir: str, asset_file_name: str, sources_digest: str, builder_id: str) -> 'PackageDigest':
        pdigest = PackageDigest(
            root_dir=root_dir,
            asset_file_name=asset_file_name,
            sources_digest=sources_digest,
            builder_id=builder_id,
        )

        pdigest._update_features()

        return pdigest

    def _update_features(self):
        # Use the asset_file_name to figure out which features are enabled.
        #
        # When name is "github-*", then it is tracking the state of a repo branch, else
        # assume it is tracking a package in dist/
        if self.asset_file_name.startswith("github-"):
            # gen_code is enabled only when doing repos branch processing.
            self._disable_package_md5()
            self._disable_dist_test()
            self._enable_gen_code()
            self._enable_ta_regtest()
        else:
            self._enable_package_md5()
            self._enable_dist_test()
            self._disable_gen_code()
            # ta_regtest must work for src.tar.gz (needed for homebrew maintainer).
            if self.asset_file_name.endswith("-src.tar.gz"):
                self._enable_ta_regtest()
            else:
                self._disable_ta_regtest()

    @staticmethod
    def from_json(root_dir: str, data: dict) -> 'PackageDigest':
        # Note: root_dir purposely not saved in JSON.

        # Exit the process on missing mandatory fields.
        # For non-mandatory fields, just initialize as if the action
        # was failed or not done.
        pdigest = PackageDigest(
            root_dir=root_dir,
            asset_file_name=data.get("asset_file_name", ""),
            sources_digest=data.get("sources_digest", ""),
            builder_id=data.get("builder_id", ""),
            built_success=data.get("built_success", ""),
            package_md5=data.get("package_md5", ""),
            gen_code_pass=data.get("gen_code_pass", ""),
            ta_regtest_pass=data.get("ta_regtest_pass", ""),
            dist_test_pass=data.get("dist_test_pass", ""),
        )

        if pdigest.asset_file_name == "" or pdigest.builder_id == "" or pdigest.sources_digest == "":
            print(f"Error: Missing mandatory field(s) in JSON: {data}")
            sys.exit(1)

        # Verify supported values.
        if pdigest.built_success not in ["True", "False"]:
            print(f"Error: Invalid value for built_success: {pdigest.built_success}")
            sys.exit(1)

        if pdigest.gen_code_pass not in ["Disabled", "Unknown", "True", "False"]:
            print(f"Error: Invalid value for gen_code_pass: {pdigest.gen_code_pass}")
            sys.exit(1)

        if pdigest.ta_regtest_pass not in ["Disabled", "Unknown", "True", "False"]:
            print(f"Error: Invalid value for ta_regtest_pass: {pdigest.ta_regtest_pass}")
            sys.exit(1)

        if pdigest.dist_test_pass not in ["Disabled", "Unknown", "True", "False"]:
            print(f"Error: Invalid value for dist_test_pass: {pdigest.dist_test_pass}")
            sys.exit(1)

        # Refresh the enabling/disabling of features.
        pdigest._update_features()

        return pdigest

    def to_json(self):
        # Note: root_dir purposely not saved in JSON.
        return {
            "asset_file_name": self.asset_file_name,
            "sources_digest": self.sources_digest,
            "builder_id": self.builder_id,
            "built_success": self.built_success,
            "package_md5": self.package_md5,
            "gen_code_pass": self.gen_code_pass,
            "ta_regtest_pass": self.ta_regtest_pass,
            "dist_test_pass": self.dist_test_pass,
        }

    def calculate_md5(self) -> str:
        if self.package_md5 == "Disabled":
            return "Disabled"
        package_file_path = path_join(self.root_dir, "dist", self.asset_file_name)
        return _calculate_md5(package_file_path)

    def clear_tests(self):
        if self.gen_code_pass != "Disabled":
            self.gen_code_pass = "Unknown"

        if self.ta_regtest_pass != "Disabled":
            self.ta_regtest_pass = "Unknown"

        if self.dist_test_pass != "Disabled":
            self.dist_test_pass = "Unknown"

    def are_all_tests_passed(self) -> bool:
        if self.gen_code_pass != "Disabled" and self.gen_code_pass != "True":
            return False
        if self.ta_regtest_pass != "Disabled" and self.ta_regtest_pass != "True":
            return False
        if self.dist_test_pass != "Disabled" and self.dist_test_pass != "True":
            return False
        return True

    def _enable_package_md5(self):
        if self.package_md5 == "Disabled":
            self.package_md5 = "Unknown"

    def _disable_package_md5(self):
        self.package_md5 = "Disabled"

    def _enable_gen_code(self):
        if self.gen_code_pass == "Disabled":
            self.gen_code_pass = "Unknown"

    def _disable_gen_code(self):
        self.gen_code_pass = "Disabled"

    def _enable_ta_regtest(self):
        if self.ta_regtest_pass == "Disabled":
            self.ta_regtest_pass = "Unknown"

    def _disable_ta_regtest(self):
        self.ta_regtest_pass = "Disabled"

    def _enable_dist_test(self):
        if self.dist_test_pass == "Disabled":
            self.dist_test_pass = "Unknown"

    def _disable_dist_test(self):
        self.dist_test_pass = "Disabled"

    def write(self):
        digests_dir = _digests_path(self.root_dir)
        os.makedirs(digests_dir, exist_ok=True)

        filepath = _asset_file_name_to_filepath(self.root_dir, self.asset_file_name)

        with open(filepath, 'w') as file:
            json.dump(self.to_json(), file, indent=4)

        # Test that the file can be read back correctly
        try:
            read_back = PackageDigest.read(self.root_dir, self.asset_file_name)
            if read_back != self:
                raise ValueError(f"Error reading back {filepath}: {read_back} != {self}")
        except Exception as e:
            raise ValueError(f"Error writing {filepath}: {e}")

    @staticmethod
    def read_or_create(root_dir: str, asset_file_name: str, sources_digest: str, builder_id: str ) -> 'PackageDigest':
        filepath = _asset_file_name_to_filepath(root_dir, asset_file_name)
        create = False

        if not os.path.exists(filepath):
            create = True
        else:
            try:
                with open(filepath, 'r') as file:
                    data = json.load(file)
                return PackageDigest.from_json(root_dir,data)
            except Exception as e:
                create = True

        if create:
            pd = PackageDigest.default(root_dir, asset_file_name, sources_digest, builder_id)
            try:
                pd.write()
                return pd
            except Exception as e:
                print(f"Error creating {filepath}: {e}")
                sys.exit(1)

    @staticmethod
    def read(root_dir: str, asset_file_name: str) -> 'PackageDigest':
        # Throws an exception if reading fails.
        filepath = _asset_file_name_to_filepath(root_dir, asset_file_name)
        create = False

        if not os.path.exists(filepath):
            raise FileNotFoundError(f"File not found: {filepath}")
        else:
            try:
                with open(filepath, 'r') as file:
                    data = json.load(file)
                return PackageDigest.from_json(root_dir,data)
            except Exception as e:
                raise ValueError(f"Error reading {filepath}: {e}")




#!/usr/bin/env python3

# Updates the website/documentation to match the latest release.
#
# Can be called directly, but more intended to be called by Github Actions.
#
import argparse
from dataclasses import dataclass
import json
import os
import re
import sys
import requests

from utilities.files import calculate_file_sha256, path_join
from utilities.common import is_brew_installed, run_command, verify_git_repo_original

try:
    from github import Github, Auth
except ImportError:
    print("This script requires the PyGithub library and it is not installed. Please do 'pip install PyGithub'")
    sys.exit(1)

def replace_version(file_path: str, version: str):
    # Replace ALL "x.y.z" pattern with "version".
    #
    # This happens only if the pattern is surrounded by characters such as "v", "[",-", "_" or whitespaces...
    # this is an heuristic logic to minimize changing unexpected text.
    with open(file_path, 'r') as f:
        content = f.read()

    # Define the regex pattern to match version numbers
    pattern = r'(?<=[v/\[\-_\s])\d+\.\d+\.\d+(?=[/\]\-_\s])'

    # Replace the matched patterns with the new version
    updated_content = re.sub(pattern, version, content)

    # Write the updated content back to the file only if there is a change
    if content != updated_content:
        with open(file_path, 'w') as f:
            f.write(updated_content)
        print(f"Updated version in {file_path} to {version}")

if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='Post-release documentation update')
    parser.add_argument('--token', help='GitHub token for authentication', required=False)
    args = parser.parse_args()

    # if an env GITHUB_TOKEN is defined... use it.
    if not args.token and 'GITHUB_TOKEN' in os.environ:
        args.token = os.getenv('GITHUB_TOKEN')
    else:
        print("Warning: No GITHUB_TOKEN env variable defined. Rate limit may fail the script.")

    if args.token:
        print("Using GitHub token for authentication")

    # if run from a fork, will exit with a message, but not an error.
    root_dir = verify_git_repo_original()

    try:
        g = Github(args.token) if args.token else Github()
        repos = g.get_repo("ta-lib/ta-lib")
        gh_version_obj = repos.get_latest_release()
        # Remove the 'v' prefix as needed
        gh_version = gh_version_obj.title.lstrip('v')

        print(f"Github release version  : {gh_version}")

        # Update version on the website installation page
        website_dir = path_join(root_dir, 'docs')
        install_md_file = path_join(website_dir, 'install.md')
        replace_version(install_md_file, gh_version)

    except Exception as e:
        print(f"Error: {e}")
        sys.exit(1)










